"""
:py:mod:`leapp.libraries.actor` represents the import location for private actor libraries that
are placed in the actor's libraries folder.


Example:
      If you actor has a libraries folder with a module.py python module, import it
      from the actor like this::

      from leapp.libraries.actor import module

"""
